/****************************************************************************
**  SIONLIB     http://www.fz-juelich.de/jsc/sionlib                       **
*****************************************************************************
**  Copyright (c) 2008-2018                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#ifndef SION_SION_DATATYPES_H
#define SION_SION_DATATYPES_H

/*!
 * \defgroup datatypes Datatypes
 * @{
 * \ingroup common
 * \brief Datatypes for use throughout the SIONlib API (deprecated)
 * \deprecated SIONlib now uses the fixed size integer types defined by the C
 *   standard libary (such as `int64_t`) in its public interface. Those types
 *   should be used instead of the ones defined here.
 */

#include <stdint.h>

/// \brief a 32 bit signed integer
/// \deprecated `int32_t` from the C standard library should be used instead
#define sion_int32 int32_t

/// \brief a 32 bit unsigned integer
/// \deprecated `uint32_t` from the C standard library should be used instead
#define sion_uint32 uint32_t

/// \brief a 64 bit signed integer
/// \deprecated `int64_t` from the C standard library should be used instead
#define sion_int64 int64_t

/// \brief a 64 bit unsigned integer
/// \deprecated `uint64_t` from the C standard library should be used instead
#define sion_uint64 uint64_t

/// @}

#endif
