#ifndef SION_SION_LIB_REP_H
#define SION_SION_LIB_REP_H
#define sion_bytes_avail_in_block lsion_bytes_avail_in_block
#define sion_bytes_avail_in_chunk lsion_bytes_avail_in_chunk
#define sion_close lsion_close
#define sion_coll_fread lsion_coll_fread
#define sion_coll_fread_mpi lsion_coll_fread_mpi
#define sion_coll_fwrite lsion_coll_fwrite
#define sion_coll_fwrite_mpi lsion_coll_fwrite_mpi
#define sion_dclose lsion_dclose
#define sion_debug_off lsion_debug_off
#define sion_debug_on lsion_debug_on
#define sion_dedup lsion_dedup
#define sion_dprintf lsion_dprintf
#define sion_dprintfp lsion_dprintfp
#define sion_dup lsion_dup
#define sion_endianness_swap_needed lsion_endianness_swap_needed
#define sion_endof_transaction_mpi lsion_endof_transaction_mpi
#define sion_ensure_free_space lsion_ensure_free_space
#define sion_fd_data lsion_fd_data
#define sion_feof lsion_feof
#define sion_file_check_par_args_free_mpi lsion_file_check_par_args_free_mpi
#define sion_file_check_par_args_init_mpi lsion_file_check_par_args_init_mpi
#define sion_file_check_par_cb_mpi lsion_file_check_par_cb_mpi
#define sion_flush lsion_flush
#define sion_fread lsion_fread
#define sion_fread_key lsion_fread_key
#define sion_fread_key_iterator_next lsion_fread_key_iterator_next
#define sion_fread_key_iterator_reset lsion_fread_key_iterator_reset
#define sion_free_io_info lsion_free_io_info
#define sion_fwrite lsion_fwrite
#define sion_fwrite_key lsion_fwrite_key
#define sion_generic_create_api lsion_generic_create_api
#define sion_generic_free_api lsion_generic_free_api
#define sion_generic_parclose lsion_generic_parclose
#define sion_generic_parclose_mapped lsion_generic_parclose_mapped
#define sion_generic_paropen lsion_generic_paropen
#define sion_generic_paropen_mapped lsion_generic_paropen_mapped
#define sion_generic_parreinit lsion_generic_parreinit
#define sion_generic_register_barrier_cb lsion_generic_register_barrier_cb
#define sion_generic_register_bcastr_cb lsion_generic_register_bcastr_cb
#define sion_generic_register_create_local_commgroup_cb lsion_generic_register_create_local_commgroup_cb
#define sion_generic_register_execute_and_scatter_cb lsion_generic_register_execute_and_scatter_cb
#define sion_generic_register_free_local_commgroup_cb lsion_generic_register_free_local_commgroup_cb
#define sion_generic_register_gather_and_execute_cb lsion_generic_register_gather_and_execute_cb
#define sion_generic_register_gatherr_cb lsion_generic_register_gatherr_cb
#define sion_generic_register_gathervr_cb lsion_generic_register_gathervr_cb
#define sion_generic_register_get_capability_cb lsion_generic_register_get_capability_cb
#define sion_generic_register_get_multi_filename_cb lsion_generic_register_get_multi_filename_cb
#define sion_generic_register_scatterr_cb lsion_generic_register_scatterr_cb
#define sion_generic_register_scattervr_cb lsion_generic_register_scattervr_cb
#define sion_get_IO_comm_mpi lsion_get_IO_comm_mpi
#define sion_get_bytes_read lsion_get_bytes_read
#define sion_get_bytes_written lsion_get_bytes_written
#define sion_get_chunksize lsion_get_chunksize
#define sion_get_current_location lsion_get_current_location
#define sion_get_current_locations lsion_get_current_locations
#define sion_get_dfile lsion_get_dfile
#define sion_get_endianness lsion_get_endianness
#define sion_get_file_endianness lsion_get_file_endianness
#define sion_get_filenumber lsion_get_filenumber
#define sion_get_fname lsion_get_fname
#define sion_get_globalrank lsion_get_globalrank
#define sion_get_io_info lsion_get_io_info
#define sion_get_io_info_buddy lsion_get_io_info_buddy
#define sion_get_io_info_by_spec lsion_get_io_info_by_spec
#define sion_get_keyval_mode lsion_get_keyval_mode
#define sion_get_locations lsion_get_locations
#define sion_get_mapping lsion_get_mapping
#define sion_get_number_of_files lsion_get_number_of_files
#define sion_get_position lsion_get_position
#define sion_get_sizeof lsion_get_sizeof
#define sion_get_version lsion_get_version
#define sion_is_serial_opened lsion_is_serial_opened
#define sion_is_thread_safe lsion_is_thread_safe
#define sion_isdebug lsion_isdebug
#define sion_key_full_scan lsion_key_full_scan
#define sion_key_get_stat lsion_key_get_stat
#define sion_key_list_iterator_next lsion_key_list_iterator_next
#define sion_key_list_iterator_reset lsion_key_list_iterator_reset
#define sion_keyval_type_to_str lsion_keyval_type_to_str
#define sion_lock_register_lock_callbacks lsion_lock_register_lock_callbacks
#define sion_lock_user_callbacks_defined lsion_lock_user_callbacks_defined
#define sion_lockdata lsion_lockdata
#define sion_mpi_options_new lsion_mpi_options_new
#define sion_mpi_options_set_buddy lsion_mpi_options_set_buddy
#define sion_mpi_options_set_buddylevel lsion_mpi_options_set_buddylevel
#define sion_mpi_options_set_chunksize lsion_mpi_options_set_chunksize
#define sion_mpi_options_set_collective lsion_mpi_options_set_collective
#define sion_mpi_options_set_collective_merge lsion_mpi_options_set_collective_merge
#define sion_mpi_options_set_collective_size lsion_mpi_options_set_collective_size
#define sion_mpi_options_set_fsblksize lsion_mpi_options_set_fsblksize
#define sion_mpi_options_set_keyval_mode lsion_mpi_options_set_keyval_mode
#define sion_mpi_options_set_multifile_communicator lsion_mpi_options_set_multifile_communicator
#define sion_mpi_options_set_multifile_number lsion_mpi_options_set_multifile_number
#define sion_open lsion_open
#define sion_open_rank lsion_open_rank
#define sion_open_with_options lsion_open_with_options
#define sion_optimize_fp_buffer lsion_optimize_fp_buffer
#define sion_options_new lsion_options_new
#define sion_options_set_chunksizes lsion_options_set_chunksizes
#define sion_options_set_fsblksize lsion_options_set_fsblksize
#define sion_options_set_keyval_mode lsion_options_set_keyval_mode
#define sion_options_set_lowlevel_api lsion_options_set_lowlevel_api
#define sion_options_set_multifile_number lsion_options_set_multifile_number
#define sion_parclose_mapped_mpi lsion_parclose_mapped_mpi
#define sion_parclose_mpi lsion_parclose_mpi
#define sion_parclose_omp lsion_parclose_omp
#define sion_parclose_ompi lsion_parclose_ompi
#define sion_paropen_comms_mpi lsion_paropen_comms_mpi
#define sion_paropen_mapped_mpi lsion_paropen_mapped_mpi
#define sion_paropen_mpi lsion_paropen_mpi
#define sion_paropen_mpi_with_options lsion_paropen_mpi_with_options
#define sion_paropen_multi_mpi lsion_paropen_multi_mpi
#define sion_paropen_omp lsion_paropen_omp
#define sion_paropen_ompi lsion_paropen_ompi
#define sion_parreinit_mpi lsion_parreinit_mpi
#define sion_seek lsion_seek
#define sion_seek_chunk lsion_seek_chunk
#define sion_seek_key lsion_seek_key
#define sion_startof_transaction_mpi lsion_startof_transaction_mpi
#define sion_swap lsion_swap
#define sion_switch_logical_file lsion_switch_logical_file
#define sion_using_hints lsion_using_hints
#endif
