/****************************************************************************
**  SIONLIB     http://www.fz-juelich.de/jsc/sionlib                       **
*****************************************************************************
**  Copyright (c) 2008-2018                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/// \cond sion_internal
/// \file

#ifndef SION_SION_HINTS_GPFS_H
#define SION_SION_HINTS_GPFS_H

/*!\brief applies hint to GPFS accessing a range in the file
 *
 * @param  fd                       file descriptor
 * @param  start                    start position of range
 * @param  length                   length of range
 * @param  iswrite                  1 for write access, 0 read access
 *
 * @return SION_SUCCESS if ok
 */
int _lsion_apply_hint_gpfs_access_range(int fd, long long start, long long length, int iswrite);

/*!\brief applies hint to GPFS freeing access to a range in the file
 *
 * @param  fd                       file descriptor
 * @param  start                    start position of range
 * @param  length                   length of range
 * @param  iswrite                  1 for write access, 0 read access
 *
 * @return SION_SUCCESS if ok
 */
int _lsion_apply_hint_gpfs_free_range(int fd, long long start, long long length, int iswrite);

#endif

/// \endcond
